from re import findall
from typing import List

from jsonpath import jsonpath
from pydantic import BaseModel
from nonebot import logger


def render(template: str, data: BaseModel) -> str:
    data = data.dict()
    # #(.*?)#
    for i in findall(r"(?<!\\)#\s*(.*?)\s*#", template):
        _ = jsonpath(data, i, use_eval=False)
        if not _ and len(_) > 0:
            logger.error(f"模板渲染失败,找不到{i}")
            res = "异常"
        else:
            res = _[0]
        template = template.replace(f"#{i}#", str(res), 1)
    return template.replace("\\#", "#")


def render_list(template: str, data: List[BaseModel]) -> str:
    return "\n".join(f"{i + 1}.{render(template, d)}" for i, d in enumerate(data))



from functools import wraps
from typing import Any, Union, Callable, Optional, List, Tuple

from arclet.alconna import Alconna
from nonebot.permission import SUPERUSER
from nonebot.typing import T_Handler
from nonebot_plugin_alconna import UniMessage, on_alconna
from nonebot_plugin_alconna.uniseg import Receipt

from .config import config, CMD_TYPE, _CM
from .depend import _raw_command


async def send(msg: Union[Any, UniMessage]) -> Optional[Receipt]:
    if not isinstance(msg, UniMessage):
        msg = UniMessage(msg)
    if str(msg) == "":
        return None
    return await msg.send(at_sender=config.nezha_at, reply_to=config.nezha_replay)


def get_cmd(cmd: CMD_TYPE) -> Optional[Tuple[bool, List[str]]]:
    if isinstance(cmd, list):
        return False, cmd
    else:
        return cmd


HELPS: List[str] = []


def wrap_cmd(cmd: CMD_TYPE, *acl_arg: Any) -> Callable:
    help_cmd = get_cmd(cmd)

    def wrapper(func: T_Handler) -> T_Handler:
        if help_cmd is None:
            return func

        @wraps(func)
        async def _catch(*args, **kwargs):
            try:
                await func(*args, **kwargs)
            except AssertionError as e:
                await send(
                    f"命令 {_raw_command()} 失败 ->\n{e}"
                )

        _ = Alconna(help_cmd[1], *acl_arg, meta=_CM)
        HELPS.append(_.get_help().strip())
        on_alconna(
            _, permission=SUPERUSER if (config.nezha_must_super_user or help_cmd[0]) else None,
            use_cmd_start=config.nezha_use_start, block=True, priority=100 - max(len(i) for i in help_cmd[1])
        ).handle()(_catch)
        return func

    return wrapper


__all__ = ["get_cmd", "send", "wrap_cmd", "HELPS"]



from abc import abstractmethod
from datetime import datetime
from typing import List, Union

from pydantic import BaseModel, root_validator

from .config import config

_n = config.nezha_decimal_places


def _rd(x):
    return round(x, _n)


class M(str):

    @classmethod
    def __get_validators__(cls):
        yield cls.transfer

    @classmethod
    @abstractmethod
    def transfer(cls, _):
        pass


class ToTime(M):
    @classmethod
    def transfer(cls, _: Union[int, float]):
        if _ < 60:
            return f"{_}s"
        elif _ < 60 * 60:
            return f"{_rd(_ / 60)}m"
        elif _ < 60 * 60 * 24:
            return f"{_rd(_ / 60 / 60)}h"
        else:
            return f"{_rd(_ / 60 / 60 / 24)}d"


class ToDatetime(M):

    @classmethod
    def transfer(cls, _: int):
        _ = datetime.fromtimestamp(_)
        return _.strftime(config.nezha_template_datetime) if _.timestamp() > 0 else "无"


class ToTotal(M):
    @classmethod
    def transfer(cls, _: int):
        # bytes to str
        if _ < 1024:
            return f"{_}B"
        elif _ < 1024 * 1024:
            return f"{_rd(_ / 1024)}K"
        elif _ < 1024 * 1024 * 1024:
            return f"{_rd(_ / 1024 / 1024)}M"
        elif _ < 1024 * 1024 * 1024 * 1024:
            return f"{_rd(_ / 1024 / 1024 / 1024)}G"
        else:
            return f"{_rd(_ / 1024 / 1024 / 1024 / 1024)}T"


class ToSpeed(M):
    @classmethod
    def transfer(cls, _: int):
        # speed to str
        if _ < 1024:
            return f"{_}B/s"
        elif _ < 1024 * 1024:
            return f"{_rd(_ / 1024)}K/s"
        elif _ < 1024 * 1024 * 1024:
            return f"{_rd(_ / 1024 / 1024)}M/s"
        elif _ < 1024 * 1024 * 1024 * 1024:
            return f"{_rd(_ / 1024 / 1024 / 1024)}G/s"
        else:
            return f"{_rd(_ / 1024 / 1024 / 1024 / 1024)}T/s"


class ToFloat(M):
    @classmethod
    def transfer(cls, _: float):
        return str(_rd(_))


class ToIP(M):
    @classmethod
    def transfer(cls, _: str):
        if _ == "":
            return "无"
        if not config.nezha_show_ip:
            return "***"
        if not config.nezha_mask_ip:
            return _
        if ":" in _:
            return _.split(":")[0] + ":***:" + _.split(":")[-1]
        return _.split(".")[0] + ".*.*." + _.split(".")[-1]


class ToOnlineStatus(M):
    @classmethod
    def transfer(cls, _: bool):
        return config.nezha_template_online_offline[0 if _ else 1]


class Host(BaseModel):
    Platform: str
    PlatformVersion: str
    CPU: List[str]
    MemTotal: ToTotal
    DiskTotal: ToTotal
    SwapTotal: ToTotal
    Arch: str
    Virtualization: str
    BootTime: ToDatetime
    CountryCode: str
    Version: str


class Status(BaseModel):
    CPU: ToFloat
    MemUsed: ToTotal
    SwapUsed: ToTotal
    DiskUsed: ToTotal
    NetInTransfer: ToTotal
    NetOutTransfer: ToTotal
    NetInSpeed: ToSpeed
    NetOutSpeed: ToSpeed
    Uptime: ToTime
    Load1: ToFloat
    Load5: ToFloat
    Load15: ToFloat
    TcpConnCount: int
    UdpConnCount: int
    ProcessCount: int


class Server(BaseModel):
    id: int
    name: str
    tag: str
    last_active: ToDatetime
    ipv4: ToIP
    ipv6: ToIP
    valid_ip: ToIP

    IfOnline: ToOnlineStatus

    @root_validator(pre=True)
    def _validator(cls, data):
        data["IfOnline"] = datetime.now().timestamp() - data["last_active"] < config.nezha_offline_time
        return data


class ServerDetails(Server):
    host: Host
    status: Status

    MemoryPercent: str
    CpuPercent: str
    DiskPercent: str

    @root_validator(pre=True)
    def _validator(cls, data):
        data["MemoryPercent"] = f'{_rd(data["status"]["MemUsed"] / data["host"]["MemTotal"] * 100)}%'
        data["CpuPercent"] = f'{_rd(data["status"]["CPU"])}%'
        data["DiskPercent"] = f'{_rd(data["status"]["DiskUsed"] / data["host"]["DiskTotal"] * 100)}%'
        data["IfOnline"] = datetime.now().timestamp() - data["last_active"] < config.nezha_offline_time
        return data



from pathlib import PurePosixPath
from typing import Literal, Any, Union, List, Optional
from urllib.parse import urlparse, urlunparse

from httpx import AsyncClient
from pydantic import parse_obj_as

from .model import Server, ServerDetails


class API:

    def __init__(self, url: str, token: str):
        self.api = url
        self.token = token
        self.headers = {
            "Authorization": token
        }

    def get_api(self, path: str) -> str:
        u = urlparse(self.api)
        return urlunparse(u._replace(path=(PurePosixPath(u.path) / path).as_posix()))  # type:ignore

    async def call_api(
            self,
            path: str, method: Literal["GET", "PUT", "POST", "DELETE", "OPTIONS", "HEAD", "PATCH", "TRACE"], **data: Any
    ) -> Any:
        data = {k: v for k, v in data.items() if v is not None}
        async with AsyncClient(headers=self.headers) as client:
            resp = await client.request(
                method, self.get_api(path), **({"params": data} if method == "GET" else {"json": data})
            )
            res = resp.json()
            assert res["code"] == 0, f"调用api失败<code:{res['code']}>: {res['message']}"
            return res.get("result")

    async def get_servers_by_tag(self, tag: Optional[str] = None) -> List[Server]:
        return parse_obj_as(List[Server], await self.call_api("/api/v1/server/list", "GET", tag=tag))

    async def get_servers_details_by_tag(self, tag: Optional[str] = None) -> List[ServerDetails]:
        return parse_obj_as(
            List[ServerDetails], await self.call_api("/api/v1/server/details", "GET", tag=tag)
        )

    async def get_servers_details_by_id(self, id_: Optional[Union[int, str]] = None) -> List[ServerDetails]:
        return parse_obj_as(
            List[ServerDetails], await self.call_api("/api/v1/server/details", "GET", id=str(id_))
        )



from typing import Optional

from nonebot.internal.matcher import current_matcher
from nonebot.internal.params import Depends
from nonebot.typing import T_State
from nonebot_plugin_alconna import ALCONNA_RESULT


def _raw_command(state: Optional[T_State] = None) -> str:
    return (state if state else current_matcher.get().state)[ALCONNA_RESULT].result.header_match.result


def RawCommand() -> str:
    return Depends(_raw_command)
